// Unreal Fighting Framework by Kevin Poretti

#include "FFState.h"

// FF includes
#include "FFStateMachineComponent.h"
#include "IFFStateAvatarInterface.h"
#include "IFFStateOwnerInterface.h"

// UE includes
#include "Components/SkeletalMeshComponent.h"

bool UFFState::CanTransition(const FFFStateContext& InStateContext)
{
    /**
     *  Check to see if avatar implements StateAvatarInterface
     *  if so then
     *      Check if the avatar is in the correct stance to perform this action
     *      Check if the state is enabled
     *      Check that all state entry conditions are met if there are any
     *
     *  Check to see if the owner implements the StateOwnerInterface
     *      Check input conditions if there are any
     *
     *  If all state entry conditions are good and at least one input condition is good then we can transition
     *  so return true otherwise return false
     */
    IFFStateAvatarInterface* SAI = Cast<IFFStateAvatarInterface>(InStateContext.Avatar);
    if(SAI)
    {
        if(!(SAI->CheckStance(StanceRequired) && SAI->CheckStateEnabled(StateType) && SAI->CheckStateEntryConditions(EntryConditions)))
        {
            return false;
        }
    }

    IFFStateOwnerInterface* SOI = Cast<IFFStateOwnerInterface>(InStateContext.Owner);
    if(SOI)
    {
        if(!SOI->CheckInputSequences(InputSequences))
        {
            return false;
        }
    }

    return OnCanTransition(InStateContext);
}


void UFFState::Enter(const FFFStateContext& InStateContext)
{
    if(InStateContext.Avatar)
    {
        USkeletalMeshComponent* SMC = InStateContext.Avatar->GetComponentByClass<USkeletalMeshComponent>();
        if(SMC && MontageToPlay)
        {
            SMC->GetAnimInstance()->Montage_Play(MontageToPlay);
        }
    }

    OnEnter(InStateContext);
}


void UFFState::Exit(const FFFStateContext& InStateContext)
{
    if(InStateContext.Avatar)
    {
        USkeletalMeshComponent* SMC = InStateContext.Avatar->GetComponentByClass<USkeletalMeshComponent>();
        if(SMC && MontageToPlay)
        {
            FAlphaBlendArgs BlendOutArgs = MontageToPlay->GetBlendOutArgs();
            SMC->GetAnimInstance()->Montage_Stop(BlendOutArgs.BlendTime, MontageToPlay);
        }
    }

    OnExit(InStateContext);
}


void UFFState::Update(float OneFrame, const FFFStateContext& InStateContext)
{
    OnUpdate(OneFrame, InStateContext);

    if(bStateHasDuration && InStateContext.Parent->GetTicksInState() >= StateDuration)
    {
        Finish(InStateContext);
    }
}


void UFFState::Finish(const FFFStateContext& InStateContext)
{
    InStateContext.Parent->GoToEntryState();
}


bool UFFState::OnCanTransition_Implementation(const FFFStateContext& InStateContext)
{
    return true;
}


UWorld* UFFState::GetWorld() const
{
	UFFStateMachineComponent* SMC = Cast<UFFStateMachineComponent>(GetOuter());
	if(SMC)
	{
		return SMC->GetWorld();
	}
	
	return nullptr;
}
